local mod, mod_name, oi = Mods.new_mod("manager")
--[[
	Manager:
		This will manage all the mod files and load them in.
		It reads out the config.json file and execute the
		included files in fixed order.
]]--

mod._mods = {}
mod._settings = {}
mod._after_run = {}

-- ####################################################################################################################
-- ##### Function #####################################################################################################
-- ####################################################################################################################	
mod.init = function(self)
	-- load config file
	self:load()
	
	-- Find all mod files and load them
	for _, mod_file in ipairs(self._get_mod_files()) do	
		local config = self._get_config_data(mod_file)
		
		if config and config.mod_name and config.version then
			config.mod_file = string.sub(mod_file, 1, -5)
			config.activated = false
			
			if self:check_config(config) then
				local stored_mod_config = self._mods[config.mod_name]
				if stored_mod_config then
					-- Check if new config has higher version number
					if self._cmp_version(stored_mod_config.version, config.version) > 0 then
						stored_mod_config = config
					end
				else
					self._mods[config.mod_name] = config
				end
			end
		end
	end
end
	
mod.load = function(self)
	local support_file = io.open("mod_loader/config/Mods.json", "r")
	if support_file then
		safe_pcall(function()
			local settings = json.decode(support_file:read("*all"))
			
			self._settings = settings
		end)
		support_file:close()
	end
end
	
mod.get = function(self, mod_name)
	if not self._mods[mod_name] then
		if mod_name and string.len(mod_name) > 0 then
			Mods.show_error(
				Localize("vmf_text_manager_no_mod_found", mod_name)
			)
		end
		return nil
	end
	return table.clone(self._mods[mod_name])
end
	
mod.run = function(self)
	local patched = true
	
	-- Keep activating mods till everything is activated or is missing requirements
	while patched do
		patched = false
		
		if self._mods then
			for mod_name, config in pairs(self._mods) do
				if config.activated == false and self:check_requirements(config) then
					--EchoConsole("[LOADED] " .. config.mod_name)
					
					-- Run language
					self:check_language(config)
					
					-- Run function
					self:check_function(config)
					
					-- Run patches
					self:check_patches(config)
					
					-- Run chat
					self:check_chat(config)
					
					config.activated = true
					patched = true
				end
			end
		end
	end
	
	-- Detect missing requirements
	for mod_name, config in pairs(self._mods) do
		if config.activated == false then
			for _, requirement_name in ipairs(config.requirement) do
				local require_mod = self._mods[requirement_name]
				if require_mod == nil or require_mod.activated == false then
					Mods.show_error(
						Localize("vmf_text_manager_missing_req", mod_name, requirement_name)
					)
					EchoConsole(
						Localize("vmf_text_manager_missing_req_advice", requirement_name)
					)
				end
			end
		end
	end
	
	-- Execute all functions when all mods has been runned
	if self._after_run then
		for _, func in ipairs(self._after_run) do
			func()
		end
	end
end
	
mod.after_run = function(self, func)
	if type(func) == "function" then
		-- Add Function
		table.insert(self._after_run, func)
	end
end
	
mod.check_config = function(self, config)
	if self._settings then
		-- If QoL is enabled
		if self._settings.whitelist then
			for _, setting in ipairs(self._settings.whitelist) do
				if setting.name == config.mod_name then
					if setting.version then
						-- Check if config has higher version number then refused mod version
						if self._cmp_version(setting.version, config.version) < 0 then
							Mods.show_error(
								Localize("vmf_text_manager_config_outdated", config.mod_file)
							)
							
							return false
						else
							return true
						end
					else
						return true
					end
				end
			end
			
			Mods.show_error(
				Localize("vmf_text_manager_config_whitelist_not_supported", config.mod_file)
			)
			
			return false
		elseif self._settings.blacklist then
			for _, setting in ipairs(self._settings.blacklist) do
				if setting.name == config.mod_name then
					if setting.version then
						-- Check if config has higher version number then refused mod version
						if self._cmp_version(setting.version, config.version) <= 0 then
							Mods.show_error(
								Localize("vmf_text_manager_config_outdated", config.mod_file)
							)
							
							return false
						end
					else
						-- refused mod
						Mods.show_error(
							Localize("vmf_text_manager_config_not_supported", config.mod_file)
						)
						
						return false
					end
				end
			end
		end
	end
	
	return true
end
	
mod.check_requirements = function(self, config)
	if config.requirement then
		for _, mod_name in ipairs(config.requirement) do
			if self._mods[mod_name] == nil or self._mods[mod_name].activated == false then 
				return false
			end
		end
	end
	
	return true
end
	
mod.check_language = function(self, config)
	if config["language"] then
		local file_path = "mods/" .. config.mod_file .. ".mod"
		
		-- Open mod file
		local zfile, zfile_err = zip.open(file_path)
		if zfile then
			for _, lang_id in ipairs(config["language"]) do
				local file_path = "language/" .. lang_id .. ".json"
				local file = zfile:open(file_path)
				
				if file then
					Mods.Language:add_file(file, lang_id)
					
					-- Close file
					file:close()
				else
					Mods.show_error(
						Localize("vmf_text_manager_could_not_open", file_path, config.mod_file)
					)
				end
			end
		end
	end
end
	
mod.check_function = function(self, config)
	if config["function"] then
		for _, function_name in ipairs(config["function"]) do
			Mods.exec(config.mod_name, function_name)
		end
	end
end
	
mod.check_patches = function(self, config)
	if config.patch then
		for _, patch_name in ipairs(config.patch) do
			Mods.exec(config.mod_name, patch_name)
		end
	end
end
	
mod.check_chat = function(self, config)
	if config.chat then
		for _, setting in ipairs(config.chat) do
			-- Activate chat command
			Mods.chat.update[#Mods.chat.update+1] = {setting.command, config.mod_name, setting.execute}
			
			--[[
				Chat can be defined in a json file. Because we can't Localize insize a json
				we must detect the description. Example: "<vmf_text_example>"
			]]-- 
			local description = setting.description
			
			-- Check description is correct to filter
			if type(description) == "string" and string.len(description) > 4 then
				-- Check string has "<" at beginning and ">" at the end
				if description:sub(1, 1) == "<" and description:sub(-1, -1) == ">" then
					description = Localize(description:sub(2, -2))
				end
			end
			
			pcall(function()
				--Add autocorrection
				SystemChat.sense.register("/" .. setting.command, description)
			end)
		end	
	end
end
	
-- Scan mods directory for mods
mod._get_mod_files = function()
	local list = {}
	
	for file in lfs.dir("./mods") do
		if string.find(file, "%.mod$") then
			table.insert(list, file)
		end
	end
	
	return list
end
	
-- Read out config data
mod._get_config_data = function(mod_file)
	-- Open zip file
	local zfile, zfile_err = zip.open("mods\\" .. mod_file)
	if zfile then
		-- Open config.json
		local config_file = zfile:open("config.json")
		
		if config_file then
			-- Read config.json
			local data = config_file:read("*all")
			
			-- Close config.json
			config_file:close()
			
			-- Convert json data to table
			local state, value = pcall(function()
				return json.decode(data)
			end)
			
			if state then
				zfile:close()
				
				return value
			else
				EchoConsole("Corrupted 'config.json' file in '" .. mod_file .. "'")
				
				Mods.show_error(
					Localize("vmf_text_manager_config_corrupted", mod_file)
				)
			end
		else
			Mods.show_error(
				Localize("vmf_text_manager_config_missing", mod_file)
			)
		end
		zfile:close()
	else
		EchoConsole(zfile_err)
	end
	
	return nil
end
	
--[[
	Compaire two versions with each other
	
	-1 = smaller then
	0 = same
	1 = bigger then
]]--
mod._cmp_version = function(old, new)
	-- Check if both versions are the same
	if old == new then
		return 0
	end
	
	-- Convery string to integer array
	local old_version = {}
	for i in string.gmatch(old, "%d+") do
		old_version[#old_version + 1] = i
	end
	
	local new_version = {}
	for i in string.gmatch(new, "%d+") do
		new_version[#new_version + 1] = i
	end
	
	-- Calculated the smallest version array list
	local length = #old_version
	if length > #new_version then length = #new_version end
	
	-- Check version number integers
	for i = 1, length do
		local n1 = tonumber(old_version[i])
		local n2 = tonumber(new_version[i])
		
		if n1 ~= n2 then
			if n1 < n2 then
				return 1
			else
				return -1
			end
		end
	end
	
	-- One of the versions has more integers in them
	if #old_version < #new_version then
		return 1
	else
		return -1
	end
end